#include <sys/event.h>
#include <sys/heap.h>
#include <bt/bt_psm.h>
#include <bt/bt_hci_api.h>
#include <stdio.h>
#include <sys/timer.h>

struct bt_psm_msg_bearer_s {
	HANDLE td;
	bt_msg_fifo* delivery_buf;
	HANDLE msg_avail;
	u_char td_kill;
	HANDLE td_killed;
};

/**
 *  \brief Representation of a higher layer service. 
 * 
 * Contains all the information needed to handle a higher layer service.
 *
 * A service is the logical endpoint that will receive and handle
 * messages obtained from a remote device. Multiple services can be registered
 * at the protocol / service multiplexor.
 * 
 * \see typedef #bt_msg (for more details about received messages)
 * \see struct bt_psm_t (for more information about the protocol / service
 * multiplexor)
 */
typedef struct _bt_psm_service_s {
	u_short psm;					///< Protocol/Service Multiplexor value
	u_short service_nr;				///< Service number uniquely defining a registered service
	BT_PSM_DATA_CB;					///< Data Reception Callback
	void* cb_arg;					///< Data Reception Callback Argument
	bt_acl_pkt_buf** free_bufs;		///< Available free buffers (opt.)
	bt_psm_msg_bearer* msg_bearer;	///< Msg bearer (opt.)
} _bt_psm_service_t;


struct bt_psm_s {
	struct btstack* bt_stack;	///< Pointer to the bt-stack to use
	_bt_psm_service_t* services;
								///< Pointer to the registered services
	u_short nr_services;		///< Maximal nr of services
	bt_acl_pkt_buf* default_buf;
								///< Pointer to the default message buffers
	u_short nr_default_bufs;
								///< Number of default message buffers
	bt_psm_msg_bearer* default_bearer;
								///< Pointer to the default message bearer
};

/**
 * \brief Returns the next free service number.
 * 
 * \param stack Pointer to the l2cap connectionless stack
 * 
 * \return
 * - next free service number
 * - #L2CAP_CL_ERR_SERVICE_PORTS_EXHAUSTED if the l2cap stack already reached
 * the maximal number of services it can handle
 */
long _bt_psm_get_next_service_nr(bt_psm_t* mux) {
	u_short i;
	_bt_psm_service_t* services = mux->services;
	for (i=0; i<mux->nr_services; i++) {
		if (services[i].psm == 0) return i;
	}
	return BT_PSM_ERR_SERVICE_PORTS_EXHAUSTED;
}

/**
 * \brief Returns the service number of the higher layer service registerd
 * on the PSM passed.
 * 
 * \param stack Pointer to the L2CAP connection-less stack
 * \param psm The Protocol/Service Multiplexer
 * 
 * \return
 * - Service number of the registered service
 * - #L2CAP_CL_ERR_NO_SERVICE_REGISTERED if the PSM passed is not in use
 */
long _bt_psm_get_service_nr_by_psm(bt_psm_t* mux, u_short psm)
{
	u_short i;
	_bt_psm_service_t* services = mux->services;
	for (i=0; i<mux->nr_services; i++) {
		if (services[i].psm == psm) return i;
	}
	return BT_PSM_ERR_NO_SERVICE_REGISTERED;
}

_bt_psm_service_t* _bt_psm_get_service_by_nr(bt_psm_t* mux, u_short service_nr)
{
	_bt_psm_service_t* service;
	// check if service nr is valid
	if (service_nr > mux->nr_services) return NULL;
	// else, check if a service is registered
	service = &mux->services[service_nr];
	if (!mux->services[service_nr].psm) return NULL;
	return service;
}

/**
 * \brief Checks if the PSM passed is valid or not.
 * 
 * \param psm Protocol/Service Multiplexer
 * 
 * \return
 * - 1 if the PSM passed is valid
 * - 0 else
 * 
 * For further details refer to the BT specification 1.2, Vol 3, Part A, p. 45.
 */
u_char _bt_psm_psm_valid(u_short psm)
{
	// TODO: implement this function
	return 1;
}

/**
 * \brief Message Bearer Thread - delivers buffered packets to a higher
 * layer service.
 * 
 * For every packet bearer there exists a packet queue and a higher layer
 * service. The packet bearer delivers the packets buffered in the packet
 * queue to the corresponding higher layer service.
 * 
 * Basically, a packet bearer waits for the event
 * #acl_pkt_bearer_t::pkt_avail (indicating that a packet has been added
 * to its packet queue), gets the next packet buffer to process from the packet
 * queue and delivers the packet to the higher layer service. This step will be
 * repeated as long as there are packets in the packet queue or the flag
 * #acl_pkt_bearer_t::td_kill is set, in which case the packet bearer
 * destroys itself and releases the resources it used back to the OS.
 * 
 * \see #acl_pkt_bearer_t
 * \see #bt_psm_pkt_fifo.c
 * \see #bt_psm_pkt_buf_destroy
 */
THREAD(MSG_BEARER, arg)
{
	bt_psm_msg_bearer* bearer;
	bt_msg* msg;
	bt_msg* free_msg;
	_bt_psm_service_t* service;
	
	// get pointer to pkt bearer
	bearer = (bt_psm_msg_bearer*)arg;
	
	while(1) {
		// if this packet bearer shall be destroyed, exit
		if (bearer->td_kill) {
			// indicate that this thread has been killed
			NutEventPostAsync(&bearer->td_killed);
			// and exit
			NutThreadExit();
		}
		// get next message from the delivery buffer
		msg = bt_msg_fifo_get_first(bearer->delivery_buf);
		// if there is a message in the queue, we deliver it
		if (msg) {
			// get pointer to service & deliver msg
			service = msg->arg;
			free_msg =
				service->data_cb(msg, msg->pload, msg->plen, service->service_nr, service->cb_arg);
			// add the msg obtained to the queue of free buffers
			free_msg->next = *(service->free_bufs);
			*(service->free_bufs) = free_msg;
		}
		// else, wait 'til a pkt buffer has to be delivered.
		else {
			NutEventWait(&bearer->msg_avail, NUT_WAIT_INFINITE);
		}
	}
}

inline bt_acl_pkt_buf** bt_psm_get_default_bufs(bt_psm_t* psmux)
{
	return &(psmux->default_buf);
}

inline bt_psm_msg_bearer* bt_psm_get_default_bearer(bt_psm_t* psmux)
{
	return psmux->default_bearer;
}

inline u_short bt_psm_get_nr_default_bufs(bt_psm_t* psmux)
{
	return psmux->nr_default_bufs;
}

short bt_psm_deliver_msg(bt_psm_t* mux,
						bt_msg* msg,
						u_char* data,
						u_short len,
						u_short psm,
						bt_acl_pkt_buf** free_buf)
{
	long service_nr;
	_bt_psm_service_t* service;
	bt_psm_msg_bearer* bearer;
	
    // if illegal psm, return error
    if (!_bt_psm_psm_valid(psm)) return BT_PSM_ERR_INVALID_PSM;
    
    // get service - return if invalid service nr
    service_nr = _bt_psm_get_service_nr_by_psm(mux, psm);
    if (service_nr < 0) return BT_PSM_ERR_NO_SERVICE_REGISTERED;
    service = &mux->services[(u_short)service_nr];
    
    // if a pkt bearer is available, add packet to delivery queue
    bearer = service->msg_bearer;
    if (bearer)
    {
    	// if there is no free packet buffer available, return error
    	if (!*(service->free_bufs)) return BT_PSM_ERR_BUF_OVERFLOW;
    	// else, get the free buffer
    	*free_buf = *(service->free_bufs);
    	*(service->free_bufs) = (*free_buf)->next;
    	(*free_buf)->next = NULL;
    	
    	/* add message to the message bearer's delivery queue */
		// store msg parameters
		msg->pload = data;
		msg->plen = len;
		msg->arg = service;
		// add msg to delivery buffer & wake up the packet bearer
		bt_msg_fifo_add_last(bearer->delivery_buf, msg);
		NutEventPostAsync(&bearer->msg_avail);
    }
    // else, deliver msg directly
    else
    {
    	*free_buf = service->data_cb(msg, data, len, service->service_nr, service->cb_arg);
    }
    return 0;
}


bt_psm_msg_bearer* bt_psm_msg_bearer_create(u_short td_size, char* td_name)
{
	bt_psm_msg_bearer* bearer;
	
	// dyn allocate packet bearer
	bearer = NutHeapAllocClear(sizeof(bt_psm_msg_bearer));
	
	// create delivery queue
	bearer->delivery_buf = bt_msg_fifo_create();
	
	// create delivery thread & return
	bearer->td = NutThreadCreate(td_name, MSG_BEARER, bearer, td_size);
	return bearer;
}

void bt_psm_msg_bearer_destroy(bt_psm_msg_bearer* bearer)
{
	// destroy delivery thread
	bearer->td_kill = 1;
	NutEventPostAsync(&bearer->msg_avail);
	NutEventWait(&bearer->td_killed, NUT_WAIT_INFINITE);

	// destroy delivery queue
	bt_msg_fifo_destroy(bearer->delivery_buf);

	// free memory
	NutHeapFree(bearer);
}

/**
 * \brief Registers an upper layer service.
 */
long bt_psm_service_register(bt_psm_t* mux,
								u_short psm,
								BT_PSM_DATA_CB,
								void* cb_arg)
{
	long service_nr;
	_bt_psm_service_t* service;
	
	// if invalid psm, return
	if (!_bt_psm_psm_valid(psm)) return BT_PSM_ERR_INVALID_PSM;
	// only continue if psm is not in use
	else if (_bt_psm_get_service_nr_by_psm(mux, psm) != BT_PSM_ERR_NO_SERVICE_REGISTERED)
		return BT_PSM_ERR_PSM_IN_USE;
	// else, get next free service control block
	service_nr = _bt_psm_get_next_service_nr(mux);
	if (service_nr < 0) return BT_PSM_ERR_SERVICE_PORTS_EXHAUSTED;
	
	// else, get pointer to service struct
	service = &mux->services[(u_short)service_nr];
	
	// set fields
	service->data_cb = data_cb;
	service->cb_arg = cb_arg;
	service->psm = psm;
	service->service_nr = service_nr;
	
	// set default message buffer & pkt bearer
	service->free_bufs = &mux->default_buf;
	service->msg_bearer = mux->default_bearer;
	return service_nr;
}

// clears the previously registered service with service nr \p service_nr
short bt_psm_service_clear(bt_psm_t* mux, u_short service_nr) {
	
	_bt_psm_service_t* service;
	
	// get pointer to the service - return if error
	if (service_nr > mux->nr_services) return BT_PSM_ERR_INVALID_SERVICE_NR;
	service = &mux->services[service_nr];
	
	// reset fields
	service->psm = 0;
	service->data_cb = NULL;
	service->cb_arg = NULL;
	return 0;
}

short bt_psm_service_set_buffers(bt_psm_t* mux,
								u_short service_nr,
								bt_acl_pkt_buf** free_bufs,
								bt_psm_msg_bearer* bearer)
{
	_bt_psm_service_t* service;
	
	// get service by nr - return if error
	service = _bt_psm_get_service_by_nr(mux, service_nr);
	if (!service) return BT_PSM_ERR_INVALID_SERVICE_NR;
	
	// set pkt queue & pkt bearer 
	service->free_bufs = free_bufs;
	service->msg_bearer = bearer;
	return 0;
}

bt_psm_t* bt_psm_init(struct btstack* bt_stack,
						u_short nr_services,
						u_short nr_bufs)
{
	bt_psm_t* bt_psm;

	// dyn. allocate psm
	bt_psm =
		NutHeapAllocClear(sizeof(bt_psm_t) + nr_services*sizeof(_bt_psm_service_t));
	bt_psm->nr_services = nr_services;
	bt_psm->services = (_bt_psm_service_t*)(((u_char*)bt_psm) + sizeof(bt_psm_t));
	bt_psm->bt_stack = bt_stack;
	
	// create default ACL packet buffer queue
	bt_psm->default_buf = bt_acl_pkt_queue_create(bt_stack, nr_bufs);
	bt_psm->nr_default_bufs = nr_bufs;
	
	// create default message bearer
	bt_psm->default_bearer = bt_psm_msg_bearer_create(1024, "psm_def");
	return bt_psm;
}
