/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

 /**
 * @file l2cap_cl.c
 *
 *\author Kevin Martin <kevmarti@tik.ee.ethz.ch>
 * 
 *\date 10-14-05
 * 
 *\brief Implementation of the connection-less l2cap layer
 *
 */
#include <string.h>
#include <sys/event.h>
#include <sys/heap.h>
#include <sys/thread.h>
#include <sys/mutex.h>
#include <sys/timer.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_hci_cmds.h>
#include <terminal/btn-terminal.h>
#include <bt/bt_msg_defs.h>
#include <bt/l2cap_cl.h>
#include <sync/sync.h>
#include <debug/toolbox.h>

#define LOG_LEVEL 0
#define LOG_CLASS 4
#include <debug/log_set.h>

typedef struct _l2cap_cl_sync_trailer_s {
	u_char psm[L2CAP_CL_PSM_FIELD_SIZE];
	u_char tstamp[SYNC_TSTAMP_LEN];
} _l2cap_cl_sync_trailer;

struct _l2cap_cl_sync_stack_s {
	struct btstack* bt_stack;
	bt_psm_t* psmux;
	u_short sync_psm;
} _l2cap_cl_sync_stack;	

bt_msg* _l2cap_cl_sync_cb(bt_msg* msg,
								u_char* data,
								u_short len,
								u_short service_nr,
								void* cb_arg)
{
	short retval;
	u_short psm;
	bt_msg* next_buf;

	// do time synchronization
	sync_process_timestamp(data + len - SYNC_TSTAMP_LEN, l2cap_cl_get_arrival_time(msg));
	
	// adjust payload size
	len -= (SYNC_TSTAMP_LEN + L2CAP_CL_PSM_FIELD_SIZE);
	
	// get original psm
	psm = data[len] | data[len + 1] << 8;
	
	DEBUG("tsync: len: %d, psm: %d\n", len, psm);
	
	// deliver the message
	retval = bt_psm_deliver_msg(_l2cap_cl_sync_stack.psmux, msg, data, len, psm, &next_buf);
	if (retval)
	{
		ERROR("tsync: delivering packet to service with psm %d! code: %d!\n",
						psm, retval);
		return msg;
	}
	return next_buf;
}

/**
 * Sends a packet over the link corresponding to the handle passed. On the
 * receiving side, the incoming packet will be handled by the service that
 * is registered at the psm specified.
 */
long l2cap_cl_sync_send_pkt(u_char* data,
							u_short len,
							bt_hci_con_handle_t hdl,
							u_short psm)
{
	long result;
	l2cap_cl_channel_t* cl_chan;
	u_char t_stamp[SYNC_TSTAMP_LEN];
	u_short send_len = len;
	u_short send_psm = psm;

	// add synchronization trailer if there is enough space
	if (len + SYNC_TSTAMP_LEN + L2CAP_CL_PSM_FIELD_SIZE <=
		bt_acl_get_max_pkt_len(_l2cap_cl_sync_stack.bt_stack) - L2CAP_CL_PKT_HDR_LEN)
	{
		send_len += SYNC_TSTAMP_LEN + L2CAP_CL_PSM_FIELD_SIZE;
		send_psm = _l2cap_cl_sync_stack.sync_psm;
	}
	
	// open l2cap cl channel
	result = l2cap_cl_channel_open(hdl, send_psm, send_len, &cl_chan);    	
	if (result) {
		ERROR("sync: open l2cap cl chan (%d)!\n", result);
		return result;
	}
		
	// write data to channel
	result = l2cap_cl_channel_write(cl_chan, data, len);
	if (result) {
		ERROR("sync: send data (%d)!\n", result);
		return result;
	}
	
	// generate timestamp and write it to channel
	if (send_psm == _l2cap_cl_sync_stack.sync_psm) {
		// send the original psm
		result = l2cap_cl_channel_write(cl_chan, (u_char*)&psm, 2);
		sync_generate_timestamp(t_stamp);
		result += l2cap_cl_channel_write(cl_chan, t_stamp, SYNC_TSTAMP_LEN);
		if (result) {
			ERROR("sending tstamp (%d)!\n", result);
			return result;
		}
	}
	
	// close channel
	l2cap_cl_channel_close(cl_chan);
	return result;
}

/**
 * Initializes the connection-less l2cap protocol stack
 */
void l2cap_cl_sync_init(struct btstack* bt_stack, bt_psm_t* mux, u_short psm)
{
	long tsync_service_nr;
	
    // store pointer to bt-stack
    _l2cap_cl_sync_stack.bt_stack = bt_stack;
    // store pointer to the protocol/service multiplexor
    _l2cap_cl_sync_stack.psmux = mux;
    _l2cap_cl_sync_stack.sync_psm = psm;

	// init synchronization layer
	rt_init();
	
	// registe time synchronization callback
	tsync_service_nr =
		bt_psm_service_register(mux, psm, _l2cap_cl_sync_cb, NULL);
	bt_psm_service_set_buffers(mux, tsync_service_nr, NULL, NULL);
}
