/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/*
 * cmds for checking the state of the point-to-point protocol layer.
 */
#include <string.h>
#include <terminal/btn-terminal.h>
#include <sys/thread.h>
#include <sys/event.h>
#include <sys/heap.h>
#include <bt/bt_hci_api.h>
#include <bt/bt_hci_cmds.h>
#include <bt/bt_psm.h>
#include <bt/l2cap_cl.h>
#include <utils/fifo_queue.h>
#include <debug/toolbox.h>

/**
 * Use this packet type to send pkts with a payload up to #BT_MAX_ACL_COM_PAYLOAD.\n
 * DM3 means three slots are used for one pkt. Smaller amount of slots can lead
 * to fragmented pkts
 */
#define HCI_PACKET_TYPE BT_HCI_PACKET_TYPE_DM3

struct bt_psm_msg_bearer_s {
	HANDLE td;
	bt_msg_fifo* delivery_buf;
	HANDLE msg_avail;
	u_char td_kill;
	HANDLE td_killed;
};

/**
 *  \brief Representation of a higher layer service. 
 * 
 * Contains all the information needed to handle a higher layer service.
 *
 * A service is the logical endpoint that will receive and handle
 * messages obtained from a remote device. Multiple services can be registered
 * at the protocol / service multiplexor.
 * 
 * \see typedef #bt_msg (for more details about received messages)
 * \see struct bt_psm_t (for more information about the protocol / service
 * multiplexor)
 */
typedef struct _bt_psm_service_s {
	u_short psm;					///< Protocol/Service Multiplexor value
	u_short service_nr;				///< Service number uniquely defining a registered service
	BT_PSM_DATA_CB;					///< Data Reception Callback
	void* cb_arg;					///< Data Reception Callback Argument
	bt_acl_pkt_buf** free_bufs;		///< Available free buffers (opt.)
	bt_psm_msg_bearer* msg_bearer;	///< Msg bearer (opt.)
} _bt_psm_service_t;


struct bt_psm_s {
	struct btstack* bt_stack;	///< Pointer to the bt-stack to use
	_bt_psm_service_t* services;
								///< Pointer to the registered services
	u_short nr_services;		///< Maximal nr of services
	bt_acl_pkt_buf* default_buf;
								///< Pointer to the default message buffers
	u_short nr_default_bufs;
								///< Number of default message buffers
	bt_psm_msg_bearer* default_bearer;
								///< Pointer to the default message bearer
};

bt_psm_t* bt_psm_cmds_psm;
char _bt_psm_cmds_str_buf[100];

void _bt_psm_cmds_print_usage(void)
{
	tprintf("psm: error: usage: services\n"
	        "                   chctxt <service_nr> <td_size>\n"
	        "                   clear <service_nr>\n");
}

void _bt_psm_cmd_show_services(void)
{
	int i;
	u_short count;
	_bt_psm_service_t* service;
	bt_msg* next_msg;
	bt_acl_pkt_buf* next_buf = NULL;
	tprintf("Services:\n");
    for (i=0; i<bt_psm_cmds_psm->nr_services; i++) {
    	// get pointer to service
    	service = &bt_psm_cmds_psm->services[i];
        if (service->psm) {
        	tprintf("service [%d]: psm: 0x%.4x, cb: %p\n", i, service->psm, service->data_cb);
	        // print out buffers
			tprintf("\tbufs: ");
			if (service->free_bufs) 
				next_buf = *(service->free_bufs);
			count = 0;
			while (next_buf) {
				tprintf("%p, ", next_buf);
				next_buf = next_buf->next;
				count++;
			}
			tprintf(" free: %d", count);
	        // if this service uses a packet bearer, print out its state
	        if (service->msg_bearer) {
	        	tprintf("\n\tcontext: %s, bufs: ",
	        		((NUTTHREADINFO*)(service->msg_bearer->td))->td_name);
	        	next_msg = service->msg_bearer->delivery_buf->first;
	        	while (next_msg) {
	        		tprintf("%p, ", next_msg);
	        		next_msg = next_msg->next;
	        	}
	        }
	        tprintf("\n");
        }
    }
}

// set buffers cmd
void _bt_psm_change_context(char* arg) {
	int res = -1;
	int service_nr, td_size;
	bt_psm_msg_bearer* bearer = NULL;
	
	// parse user input
	if (sscanf(arg, "%u%u", &service_nr, &td_size) == 2) {
		// if a msg bearer is desired, we create it, and set up the service
		if (td_size) {
			bearer = bt_psm_msg_bearer_create(td_size, "new_buf");
			res = bt_psm_service_set_buffers(bt_psm_cmds_psm,
												service_nr,
												bt_psm_cmds_psm->services[service_nr].free_bufs,
												bearer);
		}
		// if td size == 0, reset buffers and message bearer
		else {
			res = bt_psm_service_set_buffers(bt_psm_cmds_psm,
												service_nr,
												NULL,
												NULL);
		}
		if (res == BT_PSM_ERR_INVALID_SERVICE_NR)
			printf("error: invalid service nr (%d)!\n", service_nr);
		else if (res == BT_PSM_ERR_NO_SERVICE_REGISTERED)
			printf("error: no service with nr %d found!\n", service_nr);
		else {
			printf("bufset for service: %d done.\n", service_nr);
			_bt_psm_cmd_show_services();
		}
	}
	else _bt_psm_cmds_print_usage();
}

void _bt_psm_cmd_clear_service(char* arg) {
	short res;
	int service_nr;
	// parse user input
	if (sscanf(arg, "%u", &service_nr) == 1) {
		// close port
		res = bt_psm_service_clear(bt_psm_cmds_psm, service_nr);
		if (res == BT_PSM_ERR_INVALID_SERVICE_NR)
			tprintf("error: invalid service nr (%d)!\n", service_nr);
		else
			tprintf("service with nr %d cleared.\n", service_nr);
	}
	else _bt_psm_cmds_print_usage();
}

void _bt_psm_cmd_bt_psm_cmds(char* arg) {
    if (!strncmp(arg, "services", 8)) {
    	_bt_psm_cmd_show_services();
    } else if (!strncmp(arg, "clear", 5)) {
    	_bt_psm_cmd_clear_service(arg+5);
    } else if (!strncmp(arg, "chctxt", 6)) {
    	_bt_psm_change_context(arg+6);
    } else {
    	_bt_psm_cmds_print_usage();
    }
}

void bt_psm_cmds_register_cmds(void)
{
    btn_terminal_register_cmd("psm", _bt_psm_cmd_bt_psm_cmds);
}

void bt_psm_cmds_init(bt_psm_t* psmux) {
	bt_psm_cmds_psm = psmux;
}
